
// Copyright  2009. Wilbert Brants. All Rights Reserved.

package km.components {

import flash.display.*;
import flash.events.*;
import flash.filters.*;
import flash.geom.*;
import flash.media.*;
import flash.net.*;

/**
* The GUI_VideoObject class is used by KoolMoves when a video is imported.
*/
public class GUI_VideoObject extends Sprite {

	private const fBlur:BlurFilter = new BlurFilter(2, 2);
	private const YCbCr:ColorMatrixFilter = new ColorMatrixFilter([.299, .587, .114, 0, 0, -.168736, -.331264, .5, 0, 127.5, .5, -.418688, -.081312, 0, 127.5, 0, 0, 0, 1, 0]);

	private var netConnection:NetConnection;
	public var netStream:NetStream;
	public var video:Video;
	public var loop:Boolean;

	private var _postBlur:Boolean;
	private var _preBlur:Boolean;
	private var _preserveLuma:Boolean;
	private var bdAlpha:BitmapData;
	private var bdLuma:BitmapData;
	private var bmVideo:Bitmap;
	private var mapY:Array = new Array(256);
	private var mapCb:Array = new Array(256);
	private var mapCr:Array = new Array(256);
	private var mapCt:Array = new Array(256);
	private var maskMode:String;

	public function GUI_VideoObject():void {
		super();
		netConnection = new NetConnection();
		netConnection.connect(null);
		netStream = new NetStream(netConnection);
		netStream.addEventListener('asyncError', errorCatcher);
		netStream.addEventListener('ioError', errorCatcher);
		netStream.addEventListener('netStatus', netStatus);
		video = new Video();
		video.smoothing = true;
		video.attachNetStream(netStream);
		bmVideo = new Bitmap();
		addEventListener('removedFromStage', removedFromStage);
		addChild(video);
		__setup();
	}

	private function errorCatcher(e:Event):void {}

	private function netStatus(e:NetStatusEvent):void {
		if (loop && e.info.code == 'NetStream.Play.Stop') netStream.seek(0);
	}

	private function removedFromStage(e:Event):void {
		netStream.close();
		clearChromaKey();
	}

	/** @private */
	protected function __setup():void {}

	/** Sets the size of the video object.
	* @param w, width
	* @param h, height
	*/
	public function setSize(w:int = 320, h:int = 240):void {
		video.width = w;
		video.height = h;
		if (bdAlpha){
			bdAlpha = new BitmapData(w, h);
			bmVideo.bitmapData = bdAlpha.clone();
			if (_preserveLuma && _postBlur) bdLuma = new BitmapData(w, h, false);
		}
	}

	/** Starts playing a video.
	* @param url, the url of the video to play.
	*/
	public function play(url:String):void {
		netStream.play(url);
	}

	/** Clears a chroma key. */
	public function clearChromaKey():void {
		removeEventListener('enterFrame', applyChroma);
		if (contains(bmVideo)) removeChild(bmVideo);
		bdAlpha = null;
		bdLuma = null;
		addChild(video);
	}

	/** Sets a chroma key. 
	* @param keyColor, the key color.
	* @param chromaTolerance, the chroma tolerance (0 - 191) .
	* @param lumaTolerance, the luma tolerance (0 - 255) .
	* @param preBlur, blur the video before the tolerance check.
	* @param postBlur, blur the alpha channel after the tolerance check.
	* @param invertMask, when set to true, the mask is inverted.
	* @param preserveLuma, when set to true, instead of making areas transparent only the color is removed.
	*/
	public function setChromaKey(keyColor:uint = 0x00ff00, chromaTolerance:int = 60, lumaTolerance:int = 255, preBlur:Boolean = true, postBlur:Boolean = false, invertMask:Boolean = false, preserveLuma:Boolean = false):void {
		_preBlur = preBlur;
		_postBlur = postBlur;
		_preserveLuma = preserveLuma;
		if (invertMask){
			maskMode = '!=';
		} else {
			maskMode = '==';
		}
		var r:int = keyColor >> 16 & 0xff;
		var g:int = keyColor >> 8 & 0xff;
		var b:int = keyColor & 0xff;
		var Y:int = .299 * r + .587 * g + .114 * b;
		var Cb:int = 127.5 -.168736 * r -.331264 * g + .5 * b;
		var Cr:int = 127.5 + .5 * r -.418688 * g - .081312 * b;
		for (var i:int = 0; i < 256; i++){
			mapY[i] = int(Math.abs(Y - i) > lumaTolerance) << 16 | i * 0x010101 & 0xfefefe;
			mapCb[i] = Math.abs(Cb - i) >> 1;
			mapCr[i] = Math.abs(Cr - i) >> 1;
			mapCt[i] = int(i > chromaTolerance);
		}
		bdAlpha = new BitmapData(video.width, video.height);
		bmVideo.bitmapData = bdAlpha.clone();
		if (_preserveLuma && _postBlur) bdLuma = new BitmapData(bdAlpha.width, bdAlpha.height, false);
		if (contains(video)) removeChild(video);
		addChild(bmVideo);
		addEventListener('enterFrame', applyChroma, false, 0, true);
		applyChroma(new Event('set'));
	}

	private function applyChroma(e:Event):void {
		if (netStream.bufferLength || e.type == 'set'){
			bdAlpha.draw(video, video.transform.matrix);
			bmVideo.bitmapData.copyPixels(bdAlpha, bdAlpha.rect, bdAlpha.rect.topLeft);
			if (_preBlur) bdAlpha.applyFilter(bdAlpha, bdAlpha.rect, bdAlpha.rect.topLeft, fBlur);
			bdAlpha.applyFilter(bdAlpha, bdAlpha.rect, bdAlpha.rect.topLeft, YCbCr);
			bdAlpha.paletteMap(bdAlpha, bdAlpha.rect, bdAlpha.rect.topLeft, null, mapCb, mapCr);
			bdAlpha.paletteMap(bdAlpha, bdAlpha.rect, bdAlpha.rect.topLeft, mapY, null, mapCt);
			if (_postBlur){
				if (_preserveLuma) bdLuma.draw(bdAlpha);
				bdAlpha.threshold(bdAlpha, bdAlpha.rect, bdAlpha.rect.topLeft, maskMode, 0, 0, 0x010001);
				bdAlpha.applyFilter(bdAlpha, bdAlpha.rect, bdAlpha.rect.topLeft, fBlur);
				bmVideo.bitmapData.copyChannel(bdAlpha, bdAlpha.rect, bdAlpha.rect.topLeft, BitmapDataChannel.ALPHA, BitmapDataChannel.ALPHA);
				if (_preserveLuma){ bdLuma.draw(bmVideo.bitmapData); bmVideo.bitmapData.draw(bdLuma); }
			} else {
				bmVideo.bitmapData.threshold(bdAlpha, bdAlpha.rect, bdAlpha.rect.topLeft, maskMode, 0, 0, 0x010001);
				if (_preserveLuma){ bdAlpha.draw(bmVideo.bitmapData); bmVideo.bitmapData.draw(bdAlpha); }
			}
		}
	}

}
}
